"""
:py:mod:`pypara` is a Python library for

- encoding currencies and monetary value objects,
- performing monetary arithmetic and conversions, and
- running rudimentary accounting operations.

Furthermore, there are some type convenience for general use.

The source code is organised in sub-packages and sub-modules.
"""

#: Defines the version of the package.
__version__ = "0.0.25"
